/*
 *  Copyright (c) 2007 Michael Heimpold <m.heimpold@web.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2
 *  as published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program (see the file COPYING included with this
 *  distribution); if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: main.c 2 2007-01-28 22:22:49Z mhei $ */

static char *rcsrevision = "$Revision: 2 $";
#if __GNUC__ == 2
#define USE(var) static void * use_##var = (&use_##var, (void *) &var)
USE (rcsrevision);
#endif

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <errno.h>

#include "linux_avm_cpmac.h"

char *cpmac_modes[CPMAC_MODE_MAX_NO] =
  {"normal", "ata", "split", "split_ata", "all_ports", "special"};

char devname[AVM_CPMAC_MAX_DEVICE_NAME_LENGTH];
int sockfd;

void initsockfd(void) {
  if ((sockfd = socket(PF_INET, SOCK_DGRAM, 0)) == -1) {
    perror("socket");
    exit(EXIT_FAILURE);
  }
}

void prepare_ifreq(struct ifreq *ifr, void *data) {
  memset(ifr, 0, sizeof(*ifr));
  strncpy(ifr->ifr_name, devname, IFNAMSIZ);
  ifr->ifr_data = data;
}

int do_info(void) {
  struct ifreq ifr;
  struct avm_switch_info data;

  memset(&data, 0, sizeof(data));

  prepare_ifreq(&ifr, (void *)&data);

  if (ioctl(sockfd, AVM_CPMAC_IOCTL_INFO, &ifr) != 0) {
    perror("ioctl");
    exit(EXIT_FAILURE);
  }
  printf("Internal ports: %u\nExternal ports: %u\n",
         data.internal_ports, data.external_ports);
  return EXIT_SUCCESS;
}

int do_get_switch_mode(void) {
  struct avm_cpmac_ioctl_struct data;
  struct ifreq ifr;
  int ret, i;

  memset(&data, 0, sizeof(data));
  data.type = AVM_CPMAC_IOCTL_CONFIG_GET_SWITCH_MODE;

  prepare_ifreq(&ifr, (void *)&data);

  if ((ret = ioctl(sockfd, AVM_CPMAC_IOCTL_GENERIC, &ifr)) != 0) {
    switch (ret) {
      case -EIO:
        fprintf(stderr, "Driver reported: no VLAN possible.\n");
        break;
      default:
        perror("ioctl");
    }
    exit(EXIT_FAILURE);
  }
  printf("Current mode: %s\n", cpmac_modes[data.u.switch_mode.cpmac_mode - 1]);
  printf("VLAN IDs passed through WAN:");
  if (data.u.switch_mode.wan_vlan_number == 0) {
    printf(" (none)\n");
  } else {
    for (i = 0; i < data.u.switch_mode.wan_vlan_number; ++i) {
      printf(" %#x", data.u.switch_mode.wan_vlan_id[i]);
    }
    printf("\n");
  }
  return EXIT_SUCCESS;
}

int do_set_switch_mode(enum avm_cpmac_modes mode) {
  struct ifreq ifr;
  struct avm_cpmac_ioctl_struct data;
  int ret;

  memset(&data, 0, sizeof(data));
  data.type = AVM_CPMAC_IOCTL_CONFIG_SET_SWITCH_MODE;
  data.u.switch_mode.cpmac_mode = mode;

  prepare_ifreq(&ifr, (void *)&data);

  if ((ret = ioctl(sockfd, AVM_CPMAC_IOCTL_GENERIC, &ifr)) != 0) {
    switch (ret) {
      case -EIO:
        fprintf(stderr, "Driver reported: no VLAN possible.\n");
        break;
      default:
        perror("ioctl");
    }
    exit(EXIT_FAILURE);
  }
  return EXIT_SUCCESS;
}

int do_set_switch_mode_special(void) {
  struct ifreq ifr;
  struct avm_cpmac_ioctl_struct data;
  char devname[AVM_CPMAC_MAX_DEVICE_NAME_LENGTH];
  int ret, i;

  memset(&data, 0, sizeof(data));
  data.type = AVM_CPMAC_IOCTL_CONFIG_SET_SWITCH_MODE_SPECIAL;
  data.u.switch_config.number_of_devices = 2;
  data.u.switch_config.wanport           = 0xff;

  memset(&devname, 0, sizeof(devname));
  for (i = 0; i < data.u.switch_config.number_of_devices; ++i) {
    snprintf(devname, sizeof(devname), "eth%u", i);
    strncpy(data.u.switch_config.device[i].name, devname,
            sizeof(data.u.switch_config.device[i].name));
  }
  data.u.switch_config.device[0].target_mask = 0x23;
  data.u.switch_config.device[1].target_mask = 0x2c;

  prepare_ifreq(&ifr, (void *)&data);

  if ((ret = ioctl(sockfd, AVM_CPMAC_IOCTL_GENERIC, &ifr)) != 0) {
    switch (ret) {
      case -EIO:
        fprintf(stderr, "Driver reported: no VLAN possible.\n");
        break;
      default:
        perror("ioctl");
    }
    exit(EXIT_FAILURE);
  }
  return EXIT_SUCCESS;
}

int do_get_switch_mode_current(void) {
  struct ifreq ifr;
  struct avm_cpmac_ioctl_struct data;
  int ret, i;

  memset(&data, 0, sizeof(data));
  data.type = AVM_CPMAC_IOCTL_CONFIG_GET_SWITCH_MODE_CURRENT;

  prepare_ifreq(&ifr, (void *)&data);

  if ((ret = ioctl(sockfd, AVM_CPMAC_IOCTL_GENERIC, &ifr)) != 0) {
    switch (ret) {
      case -EIO:
        fprintf(stderr, "Driver reported: no VLAN possible.\n");
        break;
      default:
        perror("ioctl");
    }
    exit(EXIT_FAILURE);
  }

  printf("Devices: %u\n", data.u.switch_config.number_of_devices);
  printf("WAN is port: ");
  if (data.u.switch_config.wanport != 0xff) {
    printf("%u", data.u.switch_config.wanport);
  } else {
    printf("(none)");
  }
  printf("\n");

  for (i = 0; i < data.u.switch_config.number_of_devices; ++i)
    printf("Device %u: name=%s, portmask=0x%02x\n",
           i + 1,
           data.u.switch_config.device[i].name,
           data.u.switch_config.device[i].target_mask);
  return EXIT_SUCCESS;
}

void usage(int argc, char *argv[]) {
  fprintf(stderr, "Usage: %s [OPTIONS]... COMMAND [ARGS]\n"
  "\n"
  "Show/change VLAN configuration of avm_cpmac driver.\n"
  "\n"
  "  Options:\n"
  "    -h,--help         show this help text\n"
  "    -i,--iface        interface to operate on (default: eth0)\n"
  "\n"
  "  Commands:\n"
  "  info                    print count of internal and external ports\n"
  "\n"
  "  get_switch_mode|gsm     print current switch mode and VLAN IDs passed to WAN\n"
  "\n"
  "  set_switch_mode|ssm newmode     set switch mode to newmode\n"
  "    Implemented modes are (as stated in linux_avm_cpmac.h):\n"
  "      mode        7170             VDSL             ProfiVoIP\n"
  "      normal      eth0             eth0, wan        eth0\n"
  "      ata         eth0, wan        eth0, wan        eth0, wan\n"
  "      split       eth0-eth3        eth0-eth3, wan   eth0-eth4\n"
  "      split_ata   eth0-eth2, wan   eth0-eth2, wan   eth0-eth3, wan\n"
  "      all_ports   eth0             eth0             eth0\n"
  "      special     ???              ???              ???\n"
  "    By default mode special is a placeholder which have to be filled by issuing\n"
  "    a set_switch_mode_special command first which a custom configuration.\n"
  "\n"
  "  set_switch_mode_special|ssms    set a special switch mode configuration\n"
  "    (currently groups port 1 & 2 and port 3 & 4 to two VLANs eth0 and eth1)\n"
  "\n"
  "  get_switch_mode_current|gsmc    print current switch configuration\n"
  "\n", argv[0]);
  exit(EXIT_FAILURE);
}

int main(int argc, char *argv[]) {
  int c;
  memset(devname, 0, sizeof(devname));
  strncpy(devname, "eth0", sizeof(devname) - 1);

  while (1) {
    int option_index = 0;
    static struct option long_options[] = {
      {"help", 0, NULL, 'h'},
      {"iface", 1, NULL, 'i'},
      {0, 0, 0, 0}
    };
    c = getopt_long(argc, argv, "hi:", long_options, &option_index);
    if (c == -1)
      break;
    switch (c) {
      case 'h':
        usage(argc, argv);
        break;
      case 'i':
        strncpy(devname, optarg, sizeof(devname) - 1);
        break;
      default:
        usage(argc, argv);
    }
  }
  if (optind >= argc)
    usage(argc, argv);

  initsockfd();

  /* info */
  if (strcmp(argv[optind], "info") == 0) {
    return do_info();
  /* get_switch_mode */
  } else if (strcmp(argv[optind], "get_switch_mode") == 0 ||
             strcmp(argv[optind], "gsm") == 0) {
    return do_get_switch_mode();
  /* set_switch_mode */
  } else if (strcmp(argv[optind], "set_switch_mode") == 0 ||
             strcmp(argv[optind], "ssm") == 0) {
    if (optind + 1 >= argc) {
      fprintf(stderr, "Missing required argument newmode.\n");
      usage(argc, argv);
    }
    for (c = 0; c < CPMAC_MODE_MAX_NO - 1; ++c)
      if (strcmp(cpmac_modes[c], argv[optind + 1]) == 0)
        break;
    if (c == CPMAC_MODE_MAX_NO - 1) {
      fprintf(stderr, "Invalid switch mode: %s\n", argv[optind + 1]);
      usage(argc, argv);
    }
    fprintf(stderr, "Setting switch_mode to %u (%s).\n", c + 1, cpmac_modes[c]);
    return do_set_switch_mode(c + 1);
  /* set_switch_mode_special */
  } else if (strcmp(argv[optind], "set_switch_mode_special") == 0 ||
             strcmp(argv[optind], "ssms") == 0) {
    return do_set_switch_mode_special();
  /* get_switch_mode_current */
  } else if (strcmp(argv[optind], "get_switch_mode_current") == 0 ||
             strcmp(argv[optind], "gsmc") == 0) {
    return do_get_switch_mode_current();
  } else {
    fprintf(stderr, "Invalid command: %s", argv[optind]);
    usage(argc, argv);
  }
  exit(EXIT_SUCCESS);
}
